/**
  *  \file utf8.h
  */
#ifndef PCC_V2_UTF8_H
#define PCC_V2_UTF8_H

#include <stddef.h>
#include "cpluslib/string.h"
#include "cpluslib/types.h"

typedef uint16_t Codepage[128];

class Utf8Reader {
    const char* text;
    size_t size;

    uint8_t peek(size_t n) const;
    void advance(size_t n);
 public:
    Utf8Reader(const char* text, size_t size);
    Utf8Reader(string_t& str);

    size_t count();
    uint32_t get();
    bool hasMore() const;

    const char* getRemainingData() const;
    size_t getRemainingSize() const;

    static uint32_t makeErrorCharacter(uint8_t ch);
    static bool isErrorCharacter(uint32_t ch);
    static bool isContinuationByte(uint8_t ch);
};

string_t& appendUtf8(string_t& s, uint32_t ch);
size_t    lengthUtf8(const string_t& s);
size_t    lengthUtf8(const char* text, size_t size);
string_t  substrUtf8(const string_t& s, size_t index, size_t count);
string_t  substrUtf8(const char* text, size_t size, size_t index, size_t count);
uint32_t  charAtUtf8(const string_t& s, size_t index);
uint32_t  charAtUtf8(const char* text, size_t size, size_t index);

size_t    byteToCharPosUtf8(const string_t& s, size_t bytePos);
size_t    charToBytePosUtf8(const string_t& s, size_t charPos);

string_t  convertU16ToUtf8(const uint16_t* text, size_t count);
string_t  convertWcharToUtf8(const wchar_t* text, size_t count);
string_t  convertLatin1toUtf8(const char* text, size_t count);
string_t  convertCodepageToUtf8(const char* text, size_t count, const Codepage& cp);


/** Construct from character array.
    \param text Array
    \param size Size */
inline
Utf8Reader::Utf8Reader(const char* text, size_t size)
    : text(text), size(size)
{ }

/** Construct from string.
    \param str The string. Note that this deliberately is not a const string,
    to avoid constructing this from a temporary. */
inline
Utf8Reader::Utf8Reader(string_t& str)
    : text(str.data()), size(str.size())
{ }

/** Check whether there are more characters in this string. */
inline bool
Utf8Reader::hasMore() const
{
    return size != 0;
}

/** Get pointer to remaining data. */
inline const char*
Utf8Reader::getRemainingData() const
{
    return text;
}

/** Get number of remaining bytes. */
inline size_t
Utf8Reader::getRemainingSize() const
{
    return size;
}

/** Make an error character for a given character.
    \param ch [0x80,0xFF] (7-bit ASCII is never an error) */
inline uint32_t
Utf8Reader::makeErrorCharacter(uint8_t ch)
{
    return 0xE000 + (ch - 0x80);
}

/** Check whether the given unicode character is an error character. */
inline bool
Utf8Reader::isErrorCharacter(uint32_t ch)
{
    return (ch >= 0xE000) && (ch < 0xE080);
}

/** Check whether the given byte is a continuation byte. */
inline bool
Utf8Reader::isContinuationByte(uint8_t ch)
{
    return (ch & 0xC0) == 0x80;
}

#endif
